@@ -63,6 +63,7 @@ struct _WnckWindowPrivate
   char *session_id_utf8;
   int pid;
   int workspace;
+  gint sort_order;
 
   WnckWindowType wintype;
   
@@ -165,6 +166,9 @@ static void unqueue_update   (WnckWindow *window);
 static void queue_update     (WnckWindow *window);
 static void force_update_now (WnckWindow *window);
 
+static WnckWindow* find_last_transient_for (GList *windows,
+                                            Window xwindow);
+
 static gpointer parent_class;
 static guint signals[LAST_SIGNAL] = { 0 };
 
@@ -346,7 +350,8 @@ wnck_window_get_screen (WnckWindow *window)
 
 WnckWindow*
 _wnck_window_create (Window      xwindow,
-                     WnckScreen *screen)
+                     WnckScreen *screen,
+                     gint        sort_order)
 {
   WnckWindow *window;
 
@@ -386,6 +391,8 @@ _wnck_window_create (Window      xwindow,
                              &window->priv->y,
                              &window->priv->width,
                              &window->priv->height);
+
+  window->priv->sort_order = sort_order;
   
   window->priv->need_update_name = TRUE;
   window->priv->need_update_state = TRUE;
@@ -552,6 +559,23 @@ wnck_window_get_pid (WnckWindow *window)
   return window->priv->pid;
 }
 
+/**
+ * wnck_window_get_sort_order:
+ * @window: a #WnckWindow
+ *
+ * Gets the sort order of @window; used for ordering of the window in
+ * the tasklist.  This is defined when the window is created.
+ *
+ * Return value: sort order of @window, or G_MAXINT if none available
+ **/
+gint
+wnck_window_get_sort_order (WnckWindow *window)
+{
+  g_return_val_if_fail (WNCK_IS_WINDOW (window), G_MAXINT);
+
+  return window->priv->sort_order;
+}
+
 /**
  * wnck_window_get_window_type:
  * @window: a #WnckWindow
@@ -589,11 +613,11 @@ wnck_window_is_minimized (WnckWindow *window)
  * wnck_window_demands_attention:
  * @window: a #WnckWindow
  *
- * If the window is has the demands attention state set returns
+ * If the window has the demands attention state set returns
  * %TRUE. This state may change anytime a state_changed signal gets
  * emitted.
  *
- * Return value: %TRUE if window is minimized
+ * Return value: %TRUE if window has the demands_attention hint set
  **/
 gboolean
 wnck_window_demands_attention (WnckWindow *window)
@@ -603,6 +627,49 @@ wnck_window_demands_attention (WnckWindow *window)
   return window->priv->demands_attention;
 }
 
+/* Return whether one of the transients of @window demands attention */
+static gboolean
+transient_demands_attention (WnckWindow *window)
+{
+  GList *windows;
+  WnckWindow *transient;
+  
+  if (!WNCK_IS_WINDOW (window))
+    return FALSE;
+
+  windows = wnck_screen_get_windows_stacked (window->priv->screen);
+
+  transient = window;
+  while ((transient = find_last_transient_for (windows, transient->priv->xwindow)))
+    {
+      /* catch transient cycles */
+      if (transient == window)
+        return FALSE;
+
+      if (wnck_window_demands_attention (transient))
+        return TRUE;
+    }
+
+  return FALSE;
+}
+
+/**
+ * wnck_window_or_transient_demands_attention:
+ * @window: a #WnckWindow
+ *
+ * If the window or one of its transients has the demands attention
+ * state set returns %TRUE. This state may change anytime a
+ * state_changed signal gets emitted.
+ *
+ * Return value: %TRUE if window or one of its transients has the
+ *               demands_attention hint set
+ **/
+gboolean
+wnck_window_or_transient_demands_attention (WnckWindow *window)
+{
+  return wnck_window_demands_attention (window) || 
+         transient_demands_attention (window);
+}
 
 gboolean
 wnck_window_is_maximized_horizontally (WnckWindow *window)
@@ -785,11 +852,12 @@ wnck_window_minimize                (WnckWindow *window)
 }
 
 void
-wnck_window_unminimize              (WnckWindow *window)
+wnck_window_unminimize              (WnckWindow *window,
+                                     guint32     timestamp)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  wnck_window_activate_transient (window);
+  wnck_window_activate_transient (window, timestamp);
 }
 
 void
@@ -1025,12 +1093,14 @@ wnck_window_unpin (WnckWindow *window)
  * window manager may choose to raise @window along with focusing it.
  **/
 void
-wnck_window_activate (WnckWindow *window)
+wnck_window_activate (WnckWindow *window,
+                      guint32     timestamp)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
   _wnck_activate (WNCK_SCREEN_XSCREEN (window->priv->screen),
-		  window->priv->xwindow);
+                  window->priv->xwindow,
+                  timestamp);
 }
 
 /**
@@ -1064,12 +1134,12 @@ wnck_window_is_active (WnckWindow *window)
 gboolean
 wnck_window_is_most_recently_activated (WnckWindow *window)
 {
-  g_return_val_if_fail (WNCK_IS_WINDOW (window), FALSE);
-
   WnckWindow * current;
   WnckWindow * previous;
   WnckWindow * most_recently_activated_window;
 
+  g_return_val_if_fail (WNCK_IS_WINDOW (window), FALSE);
+
   current  = wnck_screen_get_active_window (window->priv->screen);
   previous = wnck_screen_get_previously_active_window (window->priv->screen);
 
@@ -1081,7 +1151,6 @@ wnck_window_is_most_recently_activated (WnckWindow *window)
   return (window == most_recently_activated_window);
 }
 
-
 static WnckWindow*
 find_last_transient_for (GList *windows,
                          Window xwindow)
@@ -1121,7 +1190,8 @@ find_last_transient_for (GList *windows,
  * 
  **/
 void
-wnck_window_activate_transient (WnckWindow *window)
+wnck_window_activate_transient (WnckWindow *window,
+                                guint32     timestamp)
 {
   GList *windows;
   WnckWindow *transient;
@@ -1149,16 +1219,47 @@ wnck_window_activate_transient (WnckWindow *window)
     }
 
   if (transient != NULL)
-    {
-      /* Raise but don't focus the main window */
-      XRaiseWindow (gdk_display, window->priv->xwindow);
-      /* then activate the transient */
-      wnck_window_activate (transient);
-    }
+    wnck_window_activate (transient, timestamp);
   else
+    wnck_window_activate (window, timestamp);
+}
+
+/**
+ * wnck_window_transient_is_active:
+ * @window: a #WnckWindow
+ *
+ * Return whether one of the transients of @window has focus.  This
+ * function is needed because clicking on the tasklist once will
+ * activate a transient instead of the window itself
+ * (wnck_window_activate_transient), and clicking again should
+ * minimize the window and it's transients.  (Not doing this can be
+ * especially annoying in the case of modal dialogs that don't appear
+ * in the tasklist).
+ * 
+ **/
+gboolean
+wnck_window_transient_is_active (WnckWindow *window)
+{
+  GList *windows;
+  WnckWindow *transient;
+  
+  if (!WNCK_IS_WINDOW (window))
+    return FALSE;
+
+  windows = wnck_screen_get_windows_stacked (window->priv->screen);
+
+  transient = window;
+  while ((transient = find_last_transient_for (windows, transient->priv->xwindow)))
     {
-      wnck_window_activate (window);
+      /* catch transient cycles */
+      if (transient == window)
+        return FALSE;
+
+      if (wnck_window_is_active (transient))
+        return TRUE;
     }
+
+  return FALSE;
 }
 
 static void
@@ -1795,7 +1896,8 @@ update_actions (WnckWindow *window)
                 WNCK_WINDOW_ACTION_MAXIMIZE                |
                 WNCK_WINDOW_ACTION_UNMAXIMIZE              |
                 WNCK_WINDOW_ACTION_MINIMIZE                |
-                WNCK_WINDOW_ACTION_UNMINIMIZE;
+                WNCK_WINDOW_ACTION_UNMINIMIZE              |
+                WNCK_WINDOW_ACTION_FULLSCREEN;
       return;
     }
 
@@ -1816,6 +1918,10 @@ update_actions (WnckWindow *window)
         window->priv->actions |= WNCK_WINDOW_ACTION_STICK |
                                  WNCK_WINDOW_ACTION_UNSTICK;
 
+      else if (atoms[i] == _wnck_atom_get ("_NET_WM_ACTION_MINIMIZE"))
+	window->priv->actions |= WNCK_WINDOW_ACTION_MINIMIZE   |
+	                         WNCK_WINDOW_ACTION_UNMINIMIZE;
+
       else if (atoms[i] == _wnck_atom_get ("_NET_WM_ACTION_MAXIMIZE_HORZ"))
         window->priv->actions |= WNCK_WINDOW_ACTION_MAXIMIZE_HORIZONTALLY |
                                  WNCK_WINDOW_ACTION_UNMAXIMIZE_HORIZONTALLY;
@@ -1829,6 +1935,10 @@ update_actions (WnckWindow *window)
 
       else if (atoms[i] == _wnck_atom_get ("_NET_WM_ACTION_CLOSE"))
         window->priv->actions |= WNCK_WINDOW_ACTION_CLOSE;
+
+      else if (atoms[i] == _wnck_atom_get ("_NET_WM_ACTION_FULLSCREEN"))
+        window->priv->actions |= WNCK_WINDOW_ACTION_FULLSCREEN;
+
       else
         {
           const char *name = _wnck_atom_name (atoms [i]);
@@ -1845,11 +1955,6 @@ update_actions (WnckWindow *window)
     window->priv->actions |=
         WNCK_WINDOW_ACTION_MAXIMIZE   |
         WNCK_WINDOW_ACTION_UNMAXIMIZE;
-
-  /* These are always enabled */
-  window->priv->actions |=
-        WNCK_WINDOW_ACTION_MINIMIZE   |
-        WNCK_WINDOW_ACTION_UNMINIMIZE;
 }
 
 static void
